<?php
/* vim: set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */

/**
 * TMXInfoSearcher - Search TMX info for TMO/TMS/TMN/TMU(F)/TMNF tracks
 * Created by Xymph <tm@gamers.org> based on https://api2.mania-exchange.com/
 * Based on TMXInfoFetcher (& originally https://tmuf.exchange/threadshow/619302)
 *
 * v2.0: Updated to use the current ManiaExchange API; added TMXInfo $uid,
 *       $authtime, $authscor, $vehicle; removed deprecated $length, $lbrating,
 *       $game; TMXInfo $thumburl is no longer a small version of $imageurl but
 *       the GBX thumbnail, if available; replaced paginated requests with
 *       single call at limit 500; rewrote getRecent() using antispam API call
 * v1.9: Added https support to get_file
 * v1.8: Fixed TMXInfo constructor for PHP 8.x, updated URLs to https
 * v1.7: Added Countable interface to searcher class
 * v1.6: Fixed an error checking bug
 * v1.5: Improved error reporting via $error
 * v1.4: Added User-Agent to the GET request
 * v1.3: Renamed $worldrec into $custimg to correct its meaning; fixed
 *       $replayid check
 * v1.2: Fixed TMXInfo processing false $track
 * v1.1: Optimized get_file URL parsing
 * v1.0: Initial release
 */
class TMXInfoSearcher implements Iterator,Countable {

	public $error;
	protected $tracks = array();
	private $section;
	private $site;

	/**
	 * Searches TMX for tracks matching name, author and/or environment;
	 * or search TMX for the 10 most recent tracks
	 *
	 * @param String $game
	 *        TMX section for 'TMO', 'TMS', 'TMN', 'TMU', 'TMNF'
	 * @param String $name
	 *        The track name to search for (partial, case-insensitive match)
	 * @param String $author
	 *        The track author to search for (partial, case-insensitive match)
	 * @param String $env
	 *        The environment to search for (exact case-insensitive match
	 *        from: Desert, Snow, Rally, Bay, Coast, Island, Stadium);
	 *        ignored when searching TMN or TMNF
	 * @param Boolean $recent
	 *        If true, ignore search parameters and just return 10 newest tracks
	 *        (max. one per author)
	 * @return TMXInfoSearcher
	 *        If ->valid() is false, no matching track was found;
	 *        otherwise, an iterator of TMXInfo objects for a 'foreach' loop.
	 *        Returns at most 500 tracks ($maxpage * 20) for TMNF/TMU(F),
	 *        and at most 20 tracks for the other TMX sections.
	 */
	public function __construct($game, $name, $author, $env, $recent) {

		$this->section = $game;
		switch ($game) {
		case 'TMO':
			$this->site = 'original.tm-exchange.com';
			break;
		case 'TMS':
			$this->site = 'sunrise.tm-exchange.com';
			break;
		case 'TMN':
			$this->site = 'nations.tm-exchange.com';
			$env = '';  // ignore possible environment
			break;
		case 'TMU':
			$this->site = 'tmuf.exchange';
			break;
		case 'TMNF':
			$this->site = 'tmnf.exchange';
			$env = '';  // ignore possible environment
			break;
		default:
			$this->site = '';
			$this->error = 'Unknown TMX section: ' . $game;
			return;
		}

		$fields = array(
			'TrackId',
			'TrackName',
			'UId',
			'Uploader.UserId',
			'Uploader.Name',
			'AuthorTime',
			'AuthorScore',
			'PrimaryType',
			'Environment',
			'Mood',
			'Style',
			'Routes',
			'Difficulty',
			'Car',
			'Awards',
			'Comments',
			'HasThumbnail',
			'WRReplay.ReplayId',
			'UploadedAt',
			'UpdatedAt',
			'Images',
		);
		$this->error = '';
		if ($recent) {
			$this->tracks = $this->getRecent($fields);
		} else {
			$this->tracks = $this->getList($name, $author, $env, $fields);
		}
	}  // __construct

	// define standard Iterator functions
	public function rewind() {
		reset($this->tracks);
	}
	public function current() {
		return new TMXInfo($this->section, $this->site, current($this->tracks));
	}
	public function next() {
		return new TMXInfo($this->section, $this->site, next($this->tracks));
	}
	public function key() {
		return key($this->tracks);
	}
	public function valid() {
		return (current($this->tracks) !== false);
	}
	// define standard Countable function
	public function count() {
		return count($this->tracks);
	}

	private function getRecent($fields) {

		// get 10 most recent tracks by unique author
		$url = "https://{$this->site}/api/tracks?count=10&antispam=1&fields=";
		$url .= rawurlencode(implode(',', $fields));
		$file = $this->get_file($url);
		if ($file === false) {
			$this->error = 'Connection or response error on ' . $url;
			return array();
		} else if ($file === -1) {
			$this->error = 'Timed out while reading data from ' . $url;
			return array();
		} else if ($file == '') {
			$this->error = 'No data returned from ' . $url;
			return array();
		}

		$tmx = json_decode($file);
		if ($tmx === null) {
			$this->error = 'Cannot decode recent JSON data from ' . $url;
			return array();
		}

		// check for results
		$tracks = array();
		if (isset($tmx->Results))
			$tracks = $tmx->Results;

		// return list of tracks as array of TMX objects
		return $tracks;
	}  // getRecent

	private function getList($name, $author, $env, $fields) {

		// compile search URL
		$url = "https://{$this->site}/api/tracks?count=500&fields=";
		$url .= rawurlencode(implode(',', $fields));
		if ($name != '')
			$url .= '&name=' . rawurlencode($name);
		if ($author != '')
			$url .= '&author=' . rawurlencode($author);
		switch (strtolower($env)) {
			case 'snow':
				$url .= '&environment=1';
				break;
			case 'desert':
				$url .= '&environment=2';
				break;
			case 'rally':
				$url .= '&environment=3';
				break;
			case 'island':
				$url .= '&environment=4';
				break;
			case 'coast':
				$url .= '&environment=5';
				break;
			case 'bay':
				$url .= '&environment=6';
				break;
			case 'stadium':
				$url .= '&environment=7';
				break;
		}

		// get max. 500 results
		$file = $this->get_file($url);
		if ($file === false) {
			$this->error = 'Connection or response error on ' . $url;
			return array();
		} else if ($file === -1) {
			$this->error = 'Timed out while reading data from ' . $url;
			return array();
		} else if ($file == '') {
			if ($tracks == '') {
				$this->error = 'No data returned from ' . $url;
				return array();
			}
		}

		$tmx = json_decode($file);
		if ($tmx === null) {
			$this->error = 'Cannot decode recent JSON data from ' . $url;
			return array();
		}

		// check for results
		$tracks = array();
		if (isset($tmx->Results))
			$tracks = $tmx->Results;

		// return list of tracks as array of objects
		return $tracks;
	}  // getList

	// Simple HTTP(S) Get function with timeout
	// ok: return string || error: return false || timeout: return -1
	private function get_file($url) {

		$url = parse_url($url);
		$port = isset($url['port']) ? $url['port'] :
		        ($url['scheme'] == 'https' ? 443 : 80);
		$query = isset($url['query']) ? '?' . $url['query'] : '';

		$fp = @fsockopen(($url['scheme'] == 'https' ? 'ssl://' : '') . $url['host'],
		                 $port, $errno, $errstr, 4);
		if (!$fp)
			return false;

		fwrite($fp, 'GET ' . $url['path'] . $query . " HTTP/1.0\r\n" .
		            'Host: ' . $url['host'] . "\r\n" .
		            'Content-Type: application/json' . "\r\n" .
		            'User-Agent: TMXInfoSearcher (' . PHP_OS . ")\r\n\r\n");
		stream_set_timeout($fp, 4);
		$res = '';
		$info = array();
		$info['timed_out'] = false;
		while (!feof($fp) && !$info['timed_out']) {
			$res .= fread($fp, 512);
			$info = stream_get_meta_data($fp);
		}
		fclose($fp);

		if ($info['timed_out']) {
			return -1;
		} else {
			if (substr($res, 9, 3) != '200')
				return false;
			$page = explode("\r\n\r\n", $res, 2);
			return trim($page[1]);
		}
	}  // get_file
}  // class TMXInfoSearcher


class TMXInfo {

	public $section, $site, $id, $name, $uid, $userid, $author, $authtime,
		$authscor, $type, $envir, $mood, $style, $routes, $diffic, $vehicle,
		$awards, $comments, $custimg, $uploaded, $updated,
		$pageurl, $replayid, $replayurl, $imageurl, $thumburl, $dloadurl;

	/**
	 * Returns track object with lots of data from TMX
	 *
	 * @param String $section
	 *        TMX section
	 * @param String $site
	 *        TMX site URL
	 * @param stdClass $track
	 *        The TMX track data from TMXInfoSearcher
	 * @return TMXInfo
	 */
	public function __construct($section, $site, $track) {

		$envirs = array(
			1 => 'Snow',
			2 => 'Desert',
			3 => 'Rally',
			4 => 'Island',
			5 => 'Coast',
			6 => 'Bay',
			7 => 'Stadium',
		);
		$cars = array(
			1 => 'SnowCar',
			2 => 'DesertCar',
			3 => 'RallyCar',
			4 => 'IslandCar',
			5 => 'CoastCar',
			6 => 'BayCar',
			7 => 'StadiumCar',
		);
		$moods = array(
			0 => 'Sunrise',
			1 => 'Day',
			2 => 'Sunset',
			3 => 'Night',
		);
		$types = array(
			0 => 'Race',
			1 => 'Puzzle',
			2 => 'Platform',
			3 => 'Stunts',
			4 => 'Shortcut',
			5 => 'Laps',
		);
		$routes = array(
			0 => 'Single',
			1 => 'Multiple',
			2 => 'Symmetrical',
		);
		$diffcts = array(
			0 => 'Beginner',
			1 => 'Intermediate',
			2 => 'Expert',
			3 => 'Lunatic',
		);
		$tags = array(
			0 => 'Normal',
			1 => 'Stunt',
			2 => 'Maze',
			3 => 'Offroad',
			4 => 'Laps',
			5 => 'Fullspeed',
			6 => 'LOL',
			7 => 'Tech',
			8 => 'SpeedTech',
			9 => 'RPG',
			10 => 'PressForward',
			11 => 'Trial',
			12 => 'Grass',
		);

		$this->section = $section;
		$this->site    = $site;
		if ($track) {
			$this->id       = $track->TrackId;
			$this->name     = $track->TrackName;
			$this->uid      = $track->UId;
			$this->userid   = $track->Uploader->UserId;
			$this->author   = $track->Uploader->Name;
			$this->authtime = $track->AuthorTime;
			$this->authscor = $track->AuthorScore;
			$this->type     = $types[ $track->PrimaryType ];
			$this->envir    = $envirs[ $track->Environment ];
			$this->mood     = $moods[ $track->Mood ];
			$this->style    = $tags[ $track->Style ];
			$this->routes   = $routes[ $track->Routes ];
			$this->diffic   = $diffcts[ $track->Difficulty ];
			$this->vehicle  = $cars[ $track->Car ];
			$this->awards   = $track->Awards;
			$this->comments = $track->Comments;
			$this->custimg  = !empty($track->Images);
			$this->replayid = isset($track->WRReplay->ReplayId) ? $track->WRReplay->ReplayId : 0;
			$this->uploaded = $track->UploadedAt;
			$this->updated  = $track->UpdatedAt;

			$this->pageurl  = "https://{$site}/trackshow/{$this->id}";
			$this->imageurl = $this->custimg ? "https://{$site}/trackshow/{$this->id}/image/1" : '';
			$this->thumburl = $track->HasThumbnail ? "https://{$site}/trackshow/{$this->id}/image/0" : '';
			$this->dloadurl = "https://{$site}/trackgbx/{$this->id}";

			if ($this->replayid > 0)
				$this->replayurl = "https://{$site}/recordgbx/{$this->replayid}";
			else
				$this->replayurl = '';
		}
	}  // __construct
}  // class TMXInfo

